@@ -68,6 +68,7 @@ class Poset:
     Hash is invariant under permutations.
     
     Requires external packages:
+        - numpy
         - cached_property
         - pyhash
         - pydotplus (which needs graphviz 'dot' program)
@@ -95,17 +96,16 @@ class Poset:
         several runs unless PYTHONHASHSEED is set prior to execution.
     
     Example:
-        T = Poset.total
-        V = T(2).pow(3)
+        V = Poset.from_parents([[1,2],[],[],[1]])
         V.show()
-        V = (T(2)|T(1)).downset_closure
+        V = (V|Poset.total(1)).meta_O
         V.show()
         print(V.is_distributive)
         print(V.num_f_lub_pairs)
         for f in V.iter_f_lub_pairs_bruteforce():
             V.show(f)
             print(f)
-        V.downset_closure.show()
+        V.meta_O.show()
     """
     
     def __init__(self, leq, labels=None):
@@ -120,7 +120,7 @@ class Poset:
         assert all(isinstance(l, str) for l in labels), 'labels must be strings'
         self.n = n
         self.leq = leq
-        self.labels = labels
+        self.labels = tuple(labels)
 
     # Representation methods
     
@@ -140,7 +140,7 @@ class Poset:
     
     @cached_property
     def name(self):
-        'Compact and readable representation of self'
+        'Compact and readable representation of self based on covers'
         n = self.n
         C = self.children
         topo = self.toposort
@@ -154,9 +154,10 @@ class Poset:
     
     def show(self, f=None, as_edges=False, save=None, labels=None):
         'Use graphviz to display or save self (or the endomorphism f if given)'
+        
         g = self.graphviz(f, as_edges, labels)
         png = g.create_png()
-        
+
         if save is None:
             from IPython.display import display
             from IPython.display import Image
@@ -166,7 +167,7 @@ class Poset:
             with open(save, 'wb') as f:
                 f.write(png)
         return
-    
+
     def graphviz(self, f=None, as_edges=False, labels=None):
         'Graphviz representation of self (or f if given)'
         n = self.n
@@ -234,6 +235,7 @@ class Poset:
     
     @classmethod
     def from_parents(cls, parents, labels=None):
+        'create Poset from list: parents[i] = list of parents of i'
         n = len(parents)
         children = [[] for i in range(n)]
         for ch in range(n):
@@ -243,6 +245,7 @@ class Poset:
     
     @classmethod
     def from_children(cls, children, labels=None):
+        'create Poset from list: children[i] = list of covers of i'
         n = len(children)
         child = np.zeros((n,n), dtype=bool)
         for pa in range(n):
@@ -261,10 +264,15 @@ class Poset:
     
     @classmethod
     def from_down_edges(cls, n, edges):
-        'Poset of size n that respect all edges (ancestor, descendant)'
+        'create Poset of size n respecting all given relations (ancestor, descendant)'
+        return cls.from_up_edges(n, [(j,i) for i,j in edges])
+    
+    @classmethod
+    def from_up_edges(cls, n, edges):
+        'create Poset of size n respecting all given relations (descendant, ancestor)'
         leq = np.zeros((n,n), dtype=bool)
         leq[np.diag_indices_from(leq)] = True
-        for anc, des in edges:
+        for des, anc in edges:
             leq[des, anc] = True
         leq = np.matmul(leq, leq)
         leq.flags.writeable = False
@@ -272,6 +280,7 @@ class Poset:
 
     @classmethod
     def from_lambda(cls, elems, f_leq, labels=None):
+        'create Poset with: leq[i,j] = f_leq(elems[i], elems[j])'
         m = len(elems)
         leq = np.zeros((m, m), dtype=bool)
         for i in range(m):
@@ -292,8 +301,16 @@ class Poset:
             return False # transitivity
         return True
     
+    @cached_property
+    def heights(self):
+        'Array of distance from i down to any bottom'
+        dist = self.dist
+        bottoms = self.bottoms
+        return tuple(np.min([dist[i,:] for i in bottoms], axis=0))
+
     @cached_property
     def dist(self):
+        'Matrix of shortest distance from i upwards to j'
         return self.__class__.child_to_dist(self.child)
     
     @classmethod
@@ -339,7 +356,19 @@ class Poset:
                 if indeg[v]==0:
                     q.append(v)
         len(topo)==n or self.throw('There is a cycle')
-        return topo
+        return tuple(topo)
+
+    @cached_property
+    def toporank(self):
+        return tuple(self.__class__.inverse_permutation(self.toposort))
+
+    @classmethod
+    def inverse_permutation(self, perm):
+        n = len(perm)
+        rank = [None]*n
+        for i in range(n):
+            rank[perm[i]] = i
+        return rank
     
 
     @cached_property
@@ -376,6 +405,7 @@ class Poset:
     
     @cached_property
     def lub(self):
+        'matrix of i lub j, i.e. i join j'
         n = self.n
         leq = self.leq
         lub_id = {tuple(leq[i,:]):i for i in range(n)}
@@ -404,24 +434,48 @@ class Poset:
             leq[x,glb] or self.throw(f'Not a lattice: {i} glb {j} => {glb} or {x}')
     
     @cached_property
-    def bottom(self):
-        '''bottom element of the Poset. Throws if not present'''
+    def bottoms(self):
+        'bottom elements of the poset'
         n = self.n
         nleq = self.leq.sum(axis=0)
-        zeros = [i for i in range(n) if nleq[i]==1]
-        zeros or self.throw(f'No bottom found')
-        len(zeros)==1 or self.throw(f'Multiple bottoms found: {zeros}')
-        return zeros[0]
+        return [i for i in range(n) if nleq[i]==1]
+    
+    @cached_property
+    def non_bottoms(self):
+        'non-bottom elements of the poset'
+        n = self.n
+        nleq = self.leq.sum(axis=0)
+        return [i for i in range(n) if nleq[i]>1]
+    
+    @cached_property
+    def tops(self):
+        'top elements of the poset'
+        n = self.n
+        nleq = self.leq.sum(axis=0)
+        return [i for i in range(n) if nleq[i]==n]
+
+    @cached_property
+    def non_tops(self):
+        'non-top elements of the poset'
+        n = self.n
+        nleq = self.leq.sum(axis=0)
+        return [i for i in range(n) if nleq[i]<n]
+
+    @cached_property
+    def bottom(self):
+        'unique bottom element of the Poset. Throws if not present'
+        bottoms = self.bottoms
+        bottoms or self.throw(f'No bottom found')
+        len(bottoms)==1 or self.throw(f'Multiple bottoms found: {bottoms}')
+        return bottoms[0]
     
     @cached_property
     def top(self):
-        '''top element of the Poset. Throws if not present'''
-        n = self.n
-        nleq = self.leq.sum(axis=0)
-        zeros = [i for i in range(n) if nleq[i]==n]
-        zeros or self.throw(f'No top found')
-        len(zeros)==1 or self.throw(f'Multiple bottoms found: {zeros}')
-        return zeros[0]
+        'unique top element of the Poset. Throws if not present'
+        tops = self.tops
+        tops or self.throw(f'No top found')
+        len(tops)==1 or self.throw(f'Multiple tops found: {tops}')
+        return tops[0]
     
     @cached_property
     def irreducibles(self):
@@ -532,11 +586,7 @@ class Poset:
     @cached_property
     def canonical(self):
         'representant of the equivalence class of self under reindex relation'
-        if self.is_canonical:
-            representant = self
-        else:
-            representant = self.reindex(self.canonical_index)
-        return representant
+        return self.reindex(self.canonical_index)
     
     @cached_property
     def is_canonical(self):
@@ -546,18 +596,42 @@ class Poset:
     @cached_property
     def canonical_index(self):
         'index that transforms self into the representant'
-        n = self.n
-        leq = self.leq
         h = self.hash_elems
         l = self.labels
-        def key(i):
-            return (dist[0,i], leq[:,i].sum())
-        g = sorted(range(n), key=lambda i: (leq[:,i].sum(), h[i], l[i]))
+        return self._make_index(lambda i: (h[i], l[i], i))
+    
+    def _make_index(self, key):
+        n = self.n
+        g = sorted(range(n), key=key)
         f = list(range(n))
         for i in range(n):
             f[g[i]] = i
         return f
 
+    @cached_property
+    def enumerated(self):
+        'equivalent poset with simple enumerated labels and stable order'
+        n = self.n
+        group_by = {h:[] for h in range(n)}
+        for i in range(n):
+            group_by[self.heights[i]].append(i)
+        topo = []
+        rank = [-1]*n
+        G = self.parents
+        R = self.children
+        nleq = self.leq.sum(axis=0)
+        ngeq = self.leq.sum(axis=1)
+        order = list(zip(nleq, ngeq, self.hash_elems, self.labels, range(n)))
+        def key(i):
+            t = tuple(sorted((rank[i] for i in R[i])))
+            return (t, len(G[i]), order[i])
+        for h in range(n):
+            for i in sorted(group_by[h], key=key):
+                rank[i] = len(topo)
+                topo.append(i)
+        leq = self.reindex(rank).leq
+        return self.__class__(leq, labels=None)
+
     
     # Methods for atomic changes (grow-by-one)
     
@@ -937,13 +1011,13 @@ class Poset:
     # Methods for high level relatives of self 
 
     @cached_property
-    def meta_P(self):
+    def meta_J(self):
         'subposet of join irreducibles'
         assert self.is_distributive
         return self.subgraph(self.irreducibles)
 
     @cached_property
-    def meta_L(self):
+    def meta_O(self):
         'distributive lattice of the closure of downsets of self'
         n = self.n
         leq = self.leq
@@ -994,21 +1068,22 @@ class Poset:
         return mat
 
     @cached_property
-    def meta_F(self):
+    def meta_E(self):
         'lattice of join endomorphisms of self'
         elems = list(map(tuple, self.iter_f_irreducibles_monotone_bottom()))
         labels = tuple(','.join(self._label(*f)) for f in elems)
-        return self.__class__.from_lambda(elems, self._leq_F, labels=labels)
+        return self.__class__.from_lambda(elems, self._leq_E, labels=labels)
 
-    def _leq_F(self, f, g):
+    def _leq_E(self, f, g):
         'natural order of the space of endomorphisms'
         n = self.n
         leq = self.leq
         return all(leq[f[i],g[i]] for i in range(n))
 
     @cached_property
-    def meta_JF(self):
-        'poset of functions that are join irreducibles in meta_F'
+    def meta_JE(self):
+        'poset of functions that are join irreducibles in meta_E'
+        'this is equivalent to meta_E.meta_J'
         n = self.n
         leq = self.leq
         bot = self.bottom
@@ -1016,18 +1091,16 @@ class Poset:
         f = lambda i,fi: tuple(bot if not leq[i,x] else fi for x in range(n))
         elems = [f(i, fi) for i in J for fi in J]
         labels = tuple(','.join(self._label(*f)) for f in elems)
-        return self.__class__.from_lambda(elems, self._leq_F, labels=labels)
+        return self.__class__.from_lambda(elems, self._leq_E, labels=labels)
 
 
     @cached_property
-    def meta_PP(self):
-        'poset of meta_P upside down times meta_P'
-        'with labels showing homomorphism with meta_JF'
+    def meta_JJ(self):
+        'poset of self upside down times self, i.e. (~self)*self'
+        'with labels showing homomorphism with meta_O.meta_JE'
         n = self.n
         leq = self.leq
-        bot = self.bottom
-        J = self.irreducibles
-        elems = [(i, fi) for i in J for fi in J]
+        elems = [(i, fi) for i in range(n) for fi in range(n)]
         label_of = lambda i, fi: f'f({i})={fi}'
         labels = tuple(label_of(*self._label(i,fi)) for i,fi in elems)
         def f_leq(tup_i, tup_j):
@@ -1046,7 +1119,7 @@ class Poset:
 
     def __invert__(self):
         'flip the poset upside down'
-        return self.__class__.from_children(self.parents)
+        return self.__class__.from_children(self.parents, labels=self.labels)
 
     def __add__(self, other):
         if isinstance(other, int):
@@ -1080,17 +1153,12 @@ class Poset:
     def add_poset(self, other):
         'stack other above self and connect all self.tops with all other.bottoms'
         n = self.n
-        m = other.n
-        is_top = self.leq.sum(axis=1)==1
-        is_bottom = other.leq.sum(axis=0)==1
-        tops = [i for i in range(n) if is_top[i]]
-        bottoms = [i for i in range(m) if is_bottom[i]]
         C = [
             *([j for j in Ci] for Ci in self.children),
             *([j+n for j in Ci] for Ci in other.children),
         ]
-        for i in tops:
-            for j in bottoms:
+        for i in self.tops:
+            for j in other.bottoms:
                 C[j+n].append(i)
         return self.__class__.from_children(C)
 
@@ -1098,6 +1166,7 @@ class Poset:
         'poset standard multiplication'
         n = self.n
         m = other.n
+        labels = [None]*(n*m)
         G = [[] for i in range(n*m)]
         for i in range(n):
             for j in range(m):
@@ -1105,10 +1174,11 @@ class Poset:
                     G[i+j*n].append(k+j*n)
                 for k in other.children[j]:
                     G[i+j*n].append(i+k*n)
-        return self.__class__.from_children(G)
+                labels[i+j*n] = f'({self.labels[i]},{other.labels[j]})'
+        return self.__class__.from_children(G, labels=labels)
 
     def or_poset(self, other):
-        'put other parallel to self without connections'
+        'put other at the right of self without connections'
         n = self.n
         C = [
             *([j for j in Ci] for Ci in self.children),
@@ -1119,32 +1189,25 @@ class Poset:
     def and_poset(self, other):
         'stack other above self and put self.tops * other.bottoms inbetween'
         n = self.n
-        m = other.n
-        is_top = self.leq.sum(axis=1)==1
-        is_bottom = other.leq.sum(axis=0)==1
-        tops = [i for i in range(n) if is_top[i]]
-        non_tops = [i for i in range(n) if ~is_top[i]]
-        bottoms = [i for i in range(m) if is_bottom[i]]
-        non_bottoms = [i for i in range(m) if ~is_bottom[i]]
         nodes = [
-            *((-1,i) for i in non_tops),
-            *((i,j) for i in tops for j in bottoms),
-            *((n,j) for j in non_bottoms),
+            *((-1,i) for i in self.non_tops),
+            *((i,j) for i in self.tops for j in other.bottoms),
+            *((n,j) for j in other.non_bottoms),
         ]
         C = {v:[] for v in nodes}
-        for i in non_tops:
+        for i in self.non_tops:
             for j in self.children[i]:
                 C[(-1,i)].append((-1,j))
-        for i in non_bottoms:
+        for i in other.non_bottoms:
             for j in other.parents[i]:
                 C[(n,j)].append((n,i))
-        for i in tops:
+        for i in self.tops:
             for j in self.children[i]:
-                for k in bottoms:
+                for k in other.bottoms:
                     C[(i,k)].append((-1,j))
-        for i in bottoms:
+        for i in other.bottoms:
             for j in other.parents[i]:
-                for k in tops:
+                for k in self.tops:
                     C[(n,j)].append((k,i))
         f = {node:i for i,node in enumerate(sorted(nodes))}
         children = [[] for i in range(len(f))]
@@ -1206,8 +1269,8 @@ class Poset:
     # Testing methods
     
     def _test_iters_diff(self, it1, it2):
-        'Compute set1 = set(it1)-set(it2) and set2 = set(it2)-set(it1)'
-        'Assumes that the iterators do not repeat elements'
+        '''Compute set1 = set(it1)-set(it2) and set2 = set(it2)-set(it1)
+        Assumes that the iterators do not repeat elements'''
         set1 = set()
         set2 = set()
         for x,y in zip(it1, it2):
@@ -1331,8 +1394,8 @@ class Poset:
     # Methods for serialization
     
     def to_literal(self, keys=None):
-        'Json serializable representation of self that also stores \n'
-        'some expensive cached data'
+        '''Json serializable representation of self that also stores
+        some expensive cached data'''
         out = self.__dict__.copy()
         for key, value in out.items():
             if keys is None or key in keys:
@@ -1358,6 +1421,28 @@ class Poset:
             V.__dict__[key] = value
         return V
     
+    # Methods for interactive definition of other methods
+
+    @classmethod
+    def set_method(cls, method):
+        assert hasattr(method, '__call__'), f'Not callable method: {method}'
+        setattr(cls, method.__name__, method)
+
+    @classmethod
+    def set_classmethod(cls, method):
+        assert hasattr(method, '__call__'), f'Not callable method: {method}'
+        setattr(cls, method.__name__, classmethod(method))
+
+    @classmethod
+    def set_staticmethod(cls, method):
+        assert hasattr(method, '__call__'), f'Not callable method: {method}'
+        setattr(cls, method.__name__, staticmethod(method))
+
+    @classmethod
+    def set_property(cls, method):
+        assert hasattr(method, '__call__'), f'Not callable method: {method}'
+        setattr(cls, method.__name__, property(method))
+
     # Methods related with entropy
 
     def count_antichains_bruteforce(self):
@@ -1368,7 +1453,7 @@ class Poset:
         return self.count_antichains_bruteforce()
     
     @cached_property
-    def downset_closure(self):
+    def brute_downset_closure(self):
         n = self.n
         leq = self.leq
         sets = set([frozenset()])
